@@ -26,6 +26,7 @@ import (
 	"bytes"
 	"context"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"io"
 	"os"
@@ -34,15 +35,17 @@ import (
 	"sync"
 	"syscall"
 	"time"
-	"unsafe"
 
+	"github.com/containerd/containerd/log"
 	"github.com/containerd/stargz-snapshotter/estargz"
 	commonmetrics "github.com/containerd/stargz-snapshotter/fs/metrics/common"
 	"github.com/containerd/stargz-snapshotter/fs/reader"
 	"github.com/containerd/stargz-snapshotter/fs/remote"
+	"github.com/containerd/stargz-snapshotter/metadata"
 	fusefs "github.com/hanwen/go-fuse/v2/fs"
 	"github.com/hanwen/go-fuse/v2/fuse"
 	digest "github.com/opencontainers/go-digest"
+	"github.com/sirupsen/logrus"
 	"golang.org/x/sys/unix"
 )
 
@@ -56,29 +59,90 @@ const (
 	stateDirMode      = syscall.S_IFDIR | 0500 // dr-x------
 )
 
-var opaqueXattrs = []string{"trusted.overlay.opaque", "user.overlay.opaque"}
+type OverlayOpaqueType int
 
-func newNode(layerDgst digest.Digest, r reader.Reader, blob remote.Blob) (fusefs.InodeEmbedder, error) {
-	root, ok := r.Lookup("")
-	if !ok {
-		return nil, fmt.Errorf("failed to get a TOCEntry of the root")
+const (
+	OverlayOpaqueAll OverlayOpaqueType = iota
+	OverlayOpaqueTrusted
+	OverlayOpaqueUser
+)
+
+var opaqueXattrs = map[OverlayOpaqueType][]string{
+	OverlayOpaqueAll:     {"trusted.overlay.opaque", "user.overlay.opaque"},
+	OverlayOpaqueTrusted: {"trusted.overlay.opaque"},
+	OverlayOpaqueUser:    {"user.overlay.opaque"},
+}
+
+func newNode(layerDgst digest.Digest, r reader.Reader, blob remote.Blob, baseInode uint32, opaque OverlayOpaqueType) (fusefs.InodeEmbedder, error) {
+	rootID := r.Metadata().RootID()
+	rootAttr, err := r.Metadata().GetAttr(rootID)
+	if err != nil {
+		return nil, err
 	}
+	opq, ok := opaqueXattrs[opaque]
+	if !ok {
+		return nil, fmt.Errorf("Unknown overlay opaque type")
+	}
+	ffs := &fs{
+		r:            r,
+		layerDigest:  layerDgst,
+		baseInode:    baseInode,
+		rootID:       rootID,
+		opaqueXattrs: opq,
+	}
+	ffs.s = ffs.newState(layerDgst, blob)
 	return &node{
-		r:        r,
-		e:        root,
-		s:        newState(layerDgst, blob),
-		layerSha: layerDgst,
+		id:   rootID,
+		attr: rootAttr,
+		fs:   ffs,
 	}, nil
 }
 
+// fs contains global metadata used by nodes
+type fs struct {
+	r            reader.Reader
+	s            *state
+	layerDigest  digest.Digest
+	baseInode    uint32
+	rootID       uint32
+	opaqueXattrs []string
+}
+
+func (fs *fs) inodeOfState() uint64 {
+	return (uint64(fs.baseInode) << 32) | 1 // reserved
+}
+
+func (fs *fs) inodeOfStatFile() uint64 {
+	return (uint64(fs.baseInode) << 32) | 2 // reserved
+}
+
+func (fs *fs) inodeOfID(id uint32) (uint64, error) {
+	// 0 is reserved by go-fuse 1 and 2 are reserved by the state dir
+	if id > ^uint32(0)-3 {
+		return 0, fmt.Errorf("too many inodes")
+	}
+	return (uint64(fs.baseInode) << 32) | uint64(3+id), nil
+}
+
 // node is a filesystem inode abstraction.
 type node struct {
 	fusefs.Inode
-	r        reader.Reader
-	e        *estargz.TOCEntry
-	s        *state
-	layerSha digest.Digest
-	opaque   bool // true if this node is an overlayfs opaque directory
+	fs         *fs
+	id         uint32
+	attr       metadata.Attr
+	ents       []fuse.DirEntry
+	entsCached bool
+}
+
+func (n *node) isRootNode() bool {
+	return n.id == n.fs.rootID
+}
+
+func (n *node) isOpaque() bool {
+	if _, _, err := n.fs.r.Metadata().GetChild(n.id, whiteoutOpaqueDir); err == nil {
+		return true
+	}
+	return false
 }
 
 var _ = (fusefs.InodeEmbedder)((*node)(nil))
@@ -86,47 +150,75 @@ var _ = (fusefs.InodeEmbedder)((*node)(nil))
 var _ = (fusefs.NodeReaddirer)((*node)(nil))
 
 func (n *node) Readdir(ctx context.Context) (fusefs.DirStream, syscall.Errno) {
-	// Measure how long node_readdir operation takes.
+	ents, errno := n.readdir()
+	if errno != 0 {
+		return nil, errno
+	}
+	return fusefs.NewListDirStream(ents), 0
+}
+
+func (n *node) readdir() ([]fuse.DirEntry, syscall.Errno) {
+	// Measure how long node_readdir operation takes (in microseconds).
 	start := time.Now() // set start time
-	defer commonmetrics.MeasureLatency(commonmetrics.NodeReaddir, n.layerSha, start)
+	defer commonmetrics.MeasureLatencyInMicroseconds(commonmetrics.NodeReaddir, n.fs.layerDigest, start)
+
+	if n.entsCached {
+		return n.ents, 0
+	}
+
+	isRoot := n.isRootNode()
 
 	var ents []fuse.DirEntry
-	whiteouts := map[string]*estargz.TOCEntry{}
+	whiteouts := map[string]uint32{}
 	normalEnts := map[string]bool{}
-	n.e.ForeachChild(func(baseName string, ent *estargz.TOCEntry) bool {
+	var lastErr error
+	if err := n.fs.r.Metadata().ForeachChild(n.id, func(name string, id uint32, mode os.FileMode) bool {
 
 		// We don't want to show prefetch landmarks in "/".
-		if n.e.Name == "" && (baseName == estargz.PrefetchLandmark || baseName == estargz.NoPrefetchLandmark) {
+		if isRoot && (name == estargz.PrefetchLandmark || name == estargz.NoPrefetchLandmark) {
 			return true
 		}
 
 		// We don't want to show whiteouts.
-		if strings.HasPrefix(baseName, whiteoutPrefix) {
-			if baseName == whiteoutOpaqueDir {
+		if strings.HasPrefix(name, whiteoutPrefix) {
+			if name == whiteoutOpaqueDir {
 				return true
 			}
 			// Add the overlayfs-compiant whiteout later.
-			whiteouts[baseName] = ent
+			whiteouts[name] = id
 			return true
 		}
 
 		// This is a normal entry.
-		normalEnts[baseName] = true
+		normalEnts[name] = true
+		ino, err := n.fs.inodeOfID(id)
+		if err != nil {
+			lastErr = err
+			return false
+		}
 		ents = append(ents, fuse.DirEntry{
-			Mode: modeOfEntry(ent),
-			Name: baseName,
-			Ino:  inodeOfEnt(ent),
+			Mode: fileModeToSystemMode(mode),
+			Name: name,
+			Ino:  ino,
 		})
 		return true
-	})
+	}); err != nil || lastErr != nil {
+		n.fs.s.report(fmt.Errorf("node.Readdir: err = %v; lastErr = %v", err, lastErr))
+		return nil, syscall.EIO
+	}
 
 	// Append whiteouts if no entry replaces the target entry in the lower layer.
-	for w, ent := range whiteouts {
+	for w, id := range whiteouts {
 		if !normalEnts[w[len(whiteoutPrefix):]] {
+			ino, err := n.fs.inodeOfID(id)
+			if err != nil {
+				n.fs.s.report(fmt.Errorf("node.Readdir: err = %v; lastErr = %v", err, lastErr))
+				return nil, syscall.EIO
+			}
 			ents = append(ents, fuse.DirEntry{
 				Mode: syscall.S_IFCHR,
 				Name: w[len(whiteoutPrefix):],
-				Ino:  inodeOfEnt(ent),
+				Ino:  ino,
 			})
 
 		}
@@ -136,15 +228,19 @@ func (n *node) Readdir(ctx context.Context) (fusefs.DirStream, syscall.Errno) {
 	sort.Slice(ents, func(i, j int) bool {
 		return ents[i].Name < ents[j].Name
 	})
+	n.ents, n.entsCached = ents, true // cache it
 
-	return fusefs.NewListDirStream(ents), 0
+	return ents, 0
 }
 
 var _ = (fusefs.NodeLookuper)((*node)(nil))
 
 func (n *node) Lookup(ctx context.Context, name string, out *fuse.EntryOut) (*fusefs.Inode, syscall.Errno) {
+
+	isRoot := n.isRootNode()
+
 	// We don't want to show prefetch landmarks in "/".
-	if n.e.Name == "" && (name == estargz.PrefetchLandmark || name == estargz.NoPrefetchLandmark) {
+	if isRoot && (name == estargz.PrefetchLandmark || name == estargz.NoPrefetchLandmark) {
 		return nil, syscall.ENOENT
 	}
 
@@ -154,63 +250,111 @@ func (n *node) Lookup(ctx context.Context, name string, out *fuse.EntryOut) (*fu
 	}
 
 	// state directory
-	if n.e.Name == "" && name == stateDirName {
-		return n.NewInode(ctx, n.s, stateToAttr(n.s, &out.Attr)), 0
+	if isRoot && name == stateDirName {
+		return n.NewInode(ctx, n.fs.s, n.fs.stateToAttr(&out.Attr)), 0
 	}
 
-	// lookup stargz TOCEntry
-	ce, ok := n.e.LookupChild(name)
-	if !ok {
-		// If the entry exists as a whiteout, show an overlayfs-styled whiteout node.
-		if wh, ok := n.e.LookupChild(fmt.Sprintf("%s%s", whiteoutPrefix, name)); ok {
-			return n.NewInode(ctx, &whiteout{
-				e: wh,
-			}, entryToWhAttr(wh, &out.Attr)), 0
+	// lookup on memory nodes
+	if cn := n.GetChild(name); cn != nil {
+		switch tn := cn.Operations().(type) {
+		case *node:
+			ino, err := n.fs.inodeOfID(tn.id)
+			if err != nil {
+				n.fs.s.report(fmt.Errorf("node.Lookup: %v", err))
+				return nil, syscall.EIO
+			}
+			entryToAttr(ino, tn.attr, &out.Attr)
+		case *whiteout:
+			ino, err := n.fs.inodeOfID(tn.id)
+			if err != nil {
+				n.fs.s.report(fmt.Errorf("node.Lookup: %v", err))
+				return nil, syscall.EIO
+			}
+			entryToAttr(ino, tn.attr, &out.Attr)
+		default:
+			n.fs.s.report(fmt.Errorf("node.Lookup: uknown node type detected"))
+			return nil, syscall.EIO
 		}
+		return cn, 0
+	}
+
+	// early return if this entry doesn't exist
+	if n.entsCached {
+		var found bool
+		for _, e := range n.ents {
+			if e.Name == name {
+				found = true
+			}
+		}
+		if !found {
+			return nil, syscall.ENOENT
+		}
+	}
+
+	id, ce, err := n.fs.r.Metadata().GetChild(n.id, name)
+	if err != nil {
+		// If the entry exists as a whiteout, show an overlayfs-styled whiteout node.
+		if whID, wh, err := n.fs.r.Metadata().GetChild(n.id, fmt.Sprintf("%s%s", whiteoutPrefix, name)); err == nil {
+			ino, err := n.fs.inodeOfID(whID)
+			if err != nil {
+				n.fs.s.report(fmt.Errorf("node.Lookup: %v", err))
+				return nil, syscall.EIO
+			}
+			return n.NewInode(ctx, &whiteout{
+				id:   whID,
+				fs:   n.fs,
+				attr: wh,
+			}, entryToWhAttr(ino, wh, &out.Attr)), 0
+		}
+		n.readdir() // This code path is very expensive. Cache child entries here so that the next call don't reach here.
 		return nil, syscall.ENOENT
 	}
-	var opaque bool
-	if _, ok := ce.LookupChild(whiteoutOpaqueDir); ok {
-		// This entry is an opaque directory so make it recognizable for overlayfs.
-		opaque = true
-	}
 
+	ino, err := n.fs.inodeOfID(id)
+	if err != nil {
+		n.fs.s.report(fmt.Errorf("node.Lookup: %v", err))
+		return nil, syscall.EIO
+	}
 	return n.NewInode(ctx, &node{
-		r:        n.r,
-		e:        ce,
-		s:        n.s,
-		layerSha: n.layerSha,
-		opaque:   opaque,
-	}, entryToAttr(ce, &out.Attr)), 0
+		id:   id,
+		fs:   n.fs,
+		attr: ce,
+	}, entryToAttr(ino, ce, &out.Attr)), 0
 }
 
 var _ = (fusefs.NodeOpener)((*node)(nil))
 
 func (n *node) Open(ctx context.Context, flags uint32) (fh fusefs.FileHandle, fuseFlags uint32, errno syscall.Errno) {
-	ra, err := n.r.OpenFile(n.e.Name)
+	ra, err := n.fs.r.OpenFile(n.id)
 	if err != nil {
-		n.s.report(fmt.Errorf("failed to open node: %v", err))
+		n.fs.s.report(fmt.Errorf("node.Open: %v", err))
 		return nil, 0, syscall.EIO
 	}
 	return &file{
 		n:  n,
-		e:  n.e,
 		ra: ra,
-	}, 0, 0
+	}, fuse.FOPEN_KEEP_CACHE, 0
 }
 
 var _ = (fusefs.NodeGetattrer)((*node)(nil))
 
 func (n *node) Getattr(ctx context.Context, f fusefs.FileHandle, out *fuse.AttrOut) syscall.Errno {
-	entryToAttr(n.e, &out.Attr)
+	ino, err := n.fs.inodeOfID(n.id)
+	if err != nil {
+		n.fs.s.report(fmt.Errorf("node.Getattr: %v", err))
+		return syscall.EIO
+	}
+	entryToAttr(ino, n.attr, &out.Attr)
 	return 0
 }
 
 var _ = (fusefs.NodeGetxattrer)((*node)(nil))
 
 func (n *node) Getxattr(ctx context.Context, attr string, dest []byte) (uint32, syscall.Errno) {
-	for _, opaqueXattr := range opaqueXattrs {
-		if attr == opaqueXattr && n.opaque {
+	ent := n.attr
+	opq := n.isOpaque()
+	for _, opaqueXattr := range n.fs.opaqueXattrs {
+		if attr == opaqueXattr && opq {
 			// This node is an opaque directory so give overlayfs-compliant indicator.
 			if len(dest) < len(opaqueXattrValue) {
 				return uint32(len(opaqueXattrValue)), syscall.ERANGE
@@ -218,7 +362,7 @@ func (n *node) Getxattr(ctx context.Context, attr string, dest []byte) (uint32,
 			return uint32(copy(dest, opaqueXattrValue)), 0
 		}
 	}
-	if v, ok := n.e.Xattrs[attr]; ok {
+	if v, ok := ent.Xattrs[attr]; ok {
 		if len(dest) < len(v) {
 			return uint32(len(v)), syscall.ERANGE
 		}
@@ -230,14 +374,16 @@ func (n *node) Getxattr(ctx context.Context, attr string, dest []byte) (uint32,
 var _ = (fusefs.NodeListxattrer)((*node)(nil))
 
 func (n *node) Listxattr(ctx context.Context, dest []byte) (uint32, syscall.Errno) {
+	ent := n.attr
+	opq := n.isOpaque()
 	var attrs []byte
-	if n.opaque {
+	if opq {
 		// This node is an opaque directory so add overlayfs-compliant indicator.
-		for _, opaqueXattr := range opaqueXattrs {
+		for _, opaqueXattr := range n.fs.opaqueXattrs {
 			attrs = append(attrs, []byte(opaqueXattr+"\x00")...)
 		}
 	}
-	for k := range n.e.Xattrs {
+	for k := range ent.Xattrs {
 		attrs = append(attrs, []byte(k+"\x00")...)
 	}
 	if len(dest) < len(attrs) {
@@ -249,7 +395,8 @@ func (n *node) Listxattr(ctx context.Context, dest []byte) (uint32, syscall.Errn
 var _ = (fusefs.NodeReadlinker)((*node)(nil))
 
 func (n *node) Readlink(ctx context.Context) ([]byte, syscall.Errno) {
-	return []byte(n.e.LinkName), 0
+	ent := n.attr
+	return []byte(ent.LinkName), 0
 }
 
 var _ = (fusefs.NodeStatfser)((*node)(nil))
@@ -262,18 +409,17 @@ func (n *node) Statfs(ctx context.Context, out *fuse.StatfsOut) syscall.Errno {
 // file is a file abstraction which implements file handle in go-fuse.
 type file struct {
 	n  *node
-	e  *estargz.TOCEntry
 	ra io.ReaderAt
 }
 
 var _ = (fusefs.FileReader)((*file)(nil))
 
 func (f *file) Read(ctx context.Context, dest []byte, off int64) (fuse.ReadResult, syscall.Errno) {
-	defer commonmetrics.MeasureLatency(commonmetrics.ReadOnDemand, f.n.layerSha, time.Now())   // measure time for on-demand file reads (in milliseconds)
-	defer commonmetrics.IncOperationCount(commonmetrics.OnDemandReadAccessCount, f.n.layerSha) // increment the counter for on-demand file accesses
+	defer commonmetrics.MeasureLatencyInMicroseconds(commonmetrics.ReadOnDemand, f.n.fs.layerDigest, time.Now()) // measure time for on-demand file reads (in microseconds)
+	defer commonmetrics.IncOperationCount(commonmetrics.OnDemandReadAccessCount, f.n.fs.layerDigest)             // increment the counter for on-demand file accesses
 	n, err := f.ra.ReadAt(dest, off)
 	if err != nil && err != io.EOF {
-		f.n.s.report(fmt.Errorf("failed to read node: %v", err))
+		f.n.fs.s.report(fmt.Errorf("file.Read: %v", err))
 		return nil, syscall.EIO
 	}
 	return fuse.ReadResultData(dest[:n]), 0
@@ -282,20 +428,32 @@ func (f *file) Read(ctx context.Context, dest []byte, off int64) (fuse.ReadResul
 var _ = (fusefs.FileGetattrer)((*file)(nil))
 
 func (f *file) Getattr(ctx context.Context, out *fuse.AttrOut) syscall.Errno {
-	entryToAttr(f.e, &out.Attr)
+	ino, err := f.n.fs.inodeOfID(f.n.id)
+	if err != nil {
+		f.n.fs.s.report(fmt.Errorf("file.Getattr: %v", err))
+		return syscall.EIO
+	}
+	entryToAttr(ino, f.n.attr, &out.Attr)
 	return 0
 }
 
 // whiteout is a whiteout abstraction compliant to overlayfs.
 type whiteout struct {
 	fusefs.Inode
-	e *estargz.TOCEntry
+	id   uint32
+	fs   *fs
+	attr metadata.Attr
 }
 
 var _ = (fusefs.NodeGetattrer)((*whiteout)(nil))
 
 func (w *whiteout) Getattr(ctx context.Context, f fusefs.FileHandle, out *fuse.AttrOut) syscall.Errno {
-	entryToWhAttr(w.e, &out.Attr)
+	ino, err := w.fs.inodeOfID(w.id)
+	if err != nil {
+		w.fs.s.report(fmt.Errorf("whiteout.Getattr: %v", err))
+		return syscall.EIO
+	}
+	entryToWhAttr(ino, w.attr, &out.Attr)
 	return 0
 }
 
@@ -308,7 +466,7 @@ func (w *whiteout) Statfs(ctx context.Context, out *fuse.StatfsOut) syscall.Errn
 
 // newState provides new state directory node.
 // It creates statFile at the same time to give it stable inode number.
-func newState(layerDigest digest.Digest, blob remote.Blob) *state {
+func (fs *fs) newState(layerDigest digest.Digest, blob remote.Blob) *state {
 	return &state{
 		statFile: &statFile{
 			name: layerDigest.String() + ".json",
@@ -317,7 +475,9 @@ func newState(layerDigest digest.Digest, blob remote.Blob) *state {
 				Size:   blob.Size(),
 			},
 			blob: blob,
+			fs:   fs,
 		},
+		fs: fs,
 	}
 }
 
@@ -328,6 +488,7 @@ func newState(layerDigest digest.Digest, blob remote.Blob) *state {
 type state struct {
 	fusefs.Inode
 	statFile *statFile
+	fs       *fs
 }
 
 var _ = (fusefs.NodeReaddirer)((*state)(nil))
@@ -337,7 +498,7 @@ func (s *state) Readdir(ctx context.Context) (fusefs.DirStream, syscall.Errno) {
 		{
 			Mode: statFileMode,
 			Name: s.statFile.name,
-			Ino:  inodeOfStatFile(s.statFile),
+			Ino:  s.fs.inodeOfStatFile(),
 		},
 	}), 0
 }
@@ -358,7 +519,7 @@ func (s *state) Lookup(ctx context.Context, name string, out *fuse.EntryOut) (*f
 var _ = (fusefs.NodeGetattrer)((*state)(nil))
 
 func (s *state) Getattr(ctx context.Context, f fusefs.FileHandle, out *fuse.AttrOut) syscall.Errno {
-	stateToAttr(s, &out.Attr)
+	s.fs.stateToAttr(&out.Attr)
 	return 0
 }
 
@@ -392,6 +553,7 @@ type statFile struct {
 	blob     remote.Blob
 	statJSON statJSON
 	mu       sync.Mutex
+	fs       *fs
 }
 
 var _ = (fusefs.NodeOpener)((*statFile)(nil))
@@ -430,10 +592,22 @@ func (sf *statFile) Statfs(ctx context.Context, out *fuse.StatfsOut) syscall.Err
 	return 0
 }
 
+// logContents puts the contents of statFile in the log
+// to keep that information accessible for troubleshooting.
+// The entries naming is kept to be consistend with the field naming in statJSON.
+func (sf *statFile) logContents() {
+	ctx := context.Background()
+	log.G(ctx).WithFields(logrus.Fields{
+		"digest": sf.statJSON.Digest, "size": sf.statJSON.Size,
+		"fetchedSize": sf.statJSON.FetchedSize, "fetchedPercent": sf.statJSON.FetchedPercent,
+	}).WithError(errors.New(sf.statJSON.Error)).Error("statFile error")
+}
+
 func (sf *statFile) report(err error) {
 	sf.mu.Lock()
 	defer sf.mu.Unlock()
 	sf.statJSON.Error = err.Error()
+	sf.logContents()
 }
 
 func (sf *statFile) attr(out *fuse.Attr) (fusefs.StableAttr, syscall.Errno) {
@@ -445,7 +619,7 @@ func (sf *statFile) attr(out *fuse.Attr) (fusefs.StableAttr, syscall.Errno) {
 		return fusefs.StableAttr{}, syscall.EIO
 	}
 
-	return statFileToAttr(sf, uint64(len(st)), out), 0
+	return sf.fs.statFileToAttr(uint64(len(st)), out), 0
 }
 
 func (sf *statFile) updateStatUnlocked() ([]byte, error) {
@@ -459,24 +633,21 @@ func (sf *statFile) updateStatUnlocked() ([]byte, error) {
 	return j, nil
 }
 
-// inodeOfEnt calculates the inode number which is one-to-one correspondence
-// with the TOCEntry instance.
-func inodeOfEnt(e *estargz.TOCEntry) uint64 {
-	return uint64(uintptr(unsafe.Pointer(e)))
-}
-
-// entryToAttr converts stargz's TOCEntry to go-fuse's Attr.
-func entryToAttr(e *estargz.TOCEntry, out *fuse.Attr) fusefs.StableAttr {
-	out.Ino = inodeOfEnt(e)
+// entryToAttr converts metadata.Attr to go-fuse's Attr.
+func entryToAttr(ino uint64, e metadata.Attr, out *fuse.Attr) fusefs.StableAttr {
+	out.Ino = ino
 	out.Size = uint64(e.Size)
+	if e.Mode&os.ModeSymlink != 0 {
+		out.Size = uint64(len(e.LinkName))
+	}
 	out.Blksize = blockSize
 	out.Blocks = out.Size / uint64(out.Blksize)
 	if out.Size%uint64(out.Blksize) > 0 {
 		out.Blocks++
 	}
-	mtime := e.ModTime()
+	mtime := e.ModTime
 	out.SetTimes(nil, &mtime, nil)
-	out.Mode = modeOfEntry(e)
+	out.Mode = fileModeToSystemMode(e.Mode)
 	out.Owner = fuse.Owner{Uid: uint32(e.UID), Gid: uint32(e.GID)}
 	out.Rdev = uint32(unix.Mkdev(uint32(e.DevMajor), uint32(e.DevMinor)))
 	out.Nlink = uint32(e.NumLink)
@@ -494,14 +665,13 @@ func entryToAttr(e *estargz.TOCEntry, out *fuse.Attr) fusefs.StableAttr {
 	}
 }
 
-// entryToWhAttr converts stargz's TOCEntry to go-fuse's Attr of whiteouts.
-func entryToWhAttr(e *estargz.TOCEntry, out *fuse.Attr) fusefs.StableAttr {
-	fi := e.Stat()
-	out.Ino = inodeOfEnt(e)
+// entryToWhAttr converts metadata.Attr to go-fuse's Attr of whiteouts.
+func entryToWhAttr(ino uint64, e metadata.Attr, out *fuse.Attr) fusefs.StableAttr {
+	out.Ino = ino
 	out.Size = 0
 	out.Blksize = blockSize
 	out.Blocks = 0
-	mtime := fi.ModTime()
+	mtime := e.ModTime
 	out.SetTimes(nil, &mtime, nil)
 	out.Mode = syscall.S_IFCHR
 	out.Owner = fuse.Owner{Uid: 0, Gid: 0}
@@ -518,15 +688,9 @@ func entryToWhAttr(e *estargz.TOCEntry, out *fuse.Attr) fusefs.StableAttr {
 	}
 }
 
-// inodeOfState calculates the inode number which is one-to-one correspondence
-// with the state directory instance which was created on mount.
-func inodeOfState(s *state) uint64 {
-	return uint64(uintptr(unsafe.Pointer(s)))
-}
-
 // stateToAttr converts state directory to go-fuse's Attr.
-func stateToAttr(s *state, out *fuse.Attr) fusefs.StableAttr {
-	out.Ino = inodeOfState(s)
+func (fs *fs) stateToAttr(out *fuse.Attr) fusefs.StableAttr {
+	out.Ino = fs.inodeOfState()
 	out.Size = 0
 	out.Blksize = blockSize
 	out.Blocks = 0
@@ -551,15 +715,10 @@ func stateToAttr(s *state, out *fuse.Attr) fusefs.StableAttr {
 	}
 }
 
-// inodeOfStatFile calculates the inode number which is one-to-one correspondence
-// with the stat file instance which was created on mount.
-func inodeOfStatFile(s *statFile) uint64 {
-	return uint64(uintptr(unsafe.Pointer(s)))
-}
-
 // statFileToAttr converts stat file to go-fuse's Attr.
-func statFileToAttr(sf *statFile, size uint64, out *fuse.Attr) fusefs.StableAttr {
-	out.Ino = inodeOfStatFile(sf)
+// func statFileToAttr(id uint64, sf *statFile, size uint64, out *fuse.Attr) fusefs.StableAttr {
+func (fs *fs) statFileToAttr(size uint64, out *fuse.Attr) fusefs.StableAttr {
+	out.Ino = fs.inodeOfStatFile()
 	out.Size = size
 	out.Blksize = blockSize
 	out.Blocks = out.Size / uint64(out.Blksize)
@@ -584,10 +743,7 @@ func statFileToAttr(sf *statFile, size uint64, out *fuse.Attr) fusefs.StableAttr
 	}
 }
 
-// modeOfEntry gets system's mode bits from TOCEntry
-func modeOfEntry(e *estargz.TOCEntry) uint32 {
-	m := e.Stat().Mode()
-
+func fileModeToSystemMode(m os.FileMode) uint32 {
 	// Permission bits
 	res := uint32(m & os.ModePerm)
 
