@@ -36,7 +36,7 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
 
-import os, sys, subprocess, select, random, urllib2, time, json, tempfile, shutil
+import os, sys, subprocess, select, random, urllib2, time, json, tempfile, shutil, itertools
 
 TMP_DIR = tempfile.mkdtemp()
 LEGACY_MODE = "legacy"
@@ -56,14 +56,17 @@ def exit(status):
     shutil.rmtree(TMP_DIR)
     sys.exit(status)
 
-def tmp_dir():
-    tmp_dir.nxt += 1
-    return os.path.join(TMP_DIR, str(tmp_dir.nxt))
-tmp_dir.nxt = 0
+def tmp_path():
+    tmp_path.nxt += 1
+    return os.path.join(TMP_DIR, str(tmp_path.nxt))
+tmp_path.nxt = 0
 
 def tmp_copy(src):
-    dst = tmp_dir()
-    shutil.copytree(src, dst)
+    dst = tmp_path()
+    if os.path.isdir(src):
+        shutil.copytree(src, dst)
+    else:
+        shutil.copyfile(src, dst)
     return dst
 
 def genargs_for_optimization(arg):
@@ -100,61 +103,72 @@ class Bench:
         return json.dumps(self.__dict__)
 
 class BenchRunner:
-    ECHO_HELLO = set(['alpine:3.10.2',
-                      'fedora:30',])
+    ECHO_HELLO = set(['alpine:3.15.3',
+                      'nixos/nix:2.3.12',
+                      'fedora:35',])
 
-    CMD_ARG_WAIT = {'rethinkdb:2.3.6': RunArgs(waitline='Server ready'),
+    CMD_ARG_WAIT = {'rethinkdb:2.4.1': RunArgs(waitline='Server ready'),
                     'glassfish:4.1-jdk8': RunArgs(waitline='Running GlassFish'),
-                    'drupal:8.7.6': RunArgs(waitline='apache2 -D FOREGROUND'),
+                    'drupal:9.3.9': RunArgs(waitline='apache2 -D FOREGROUND'),
                     'jenkins:2.60.3': RunArgs(waitline='Jenkins is fully up and running'),
-                    'redis:5.0.5': RunArgs(waitline='Ready to accept connections'),
-                    'tomcat:10.0.0-jdk15-openjdk-buster': RunArgs(waitline='Server startup'),
-                    'postgres:13.1': RunArgs(waitline='database system is ready to accept connections',
+                    'redis:6.2.6': RunArgs(waitline='Ready to accept connections'),
+                    'tomcat:10.1.0-jdk17-openjdk-bullseye': RunArgs(waitline='Server startup'),
+                    'postgres:14.2': RunArgs(waitline='database system is ready to accept connections',
                                              env={'POSTGRES_PASSWORD': 'abc'}),
-                    'mariadb:10.5': RunArgs(waitline='mysqld: ready for connections',
+                    'mariadb:10.7.3': RunArgs(waitline='mysqld: ready for connections',
                                             env={'MYSQL_ROOT_PASSWORD': 'abc'}),
-                    'wordpress:5.7': RunArgs(waitline='apache2 -D FOREGROUND'),
+                    'wordpress:5.9.2': RunArgs(waitline='apache2 -D FOREGROUND'),
+                    'php:8.1.4-apache-bullseye': RunArgs(waitline='apache2 -D FOREGROUND'),
+                    'rabbitmq:3.9.14': RunArgs(waitline='Server startup complete'),
+                    'elasticsearch:8.1.1': RunArgs(waitline='"started"',
+                                                    mount=[('elasticsearch/elasticsearch.yml', '/usr/share/elasticsearch/config/elasticsearch.yml')]),
     }
 
-    CMD_STDIN = {'php:7.3.8':  RunArgs(stdin='php -r "echo \\\"hello\\n\\\";"; exit\n'),
-                 'gcc:10.2.0': RunArgs(stdin='cd /src; gcc main.c; ./a.out; exit\n',
+    CMD_STDIN = {'php:8.1.4':  RunArgs(stdin='php -r "echo \\\"hello\\n\\\";"; exit\n'),
+                 'gcc:11.2.0': RunArgs(stdin='cd /src; gcc main.c; ./a.out; exit\n',
                                 mount=[('gcc', '/src')]),
-                 'golang:1.12.9': RunArgs(stdin='cd /go/src; go run main.go; exit\n',
+                 'golang:1.18': RunArgs(stdin='cd /go/src; go run main.go; exit\n',
                                    mount=[('go', '/go/src')]),
-                 'jruby:9.2.8.0': RunArgs(stdin='jruby -e "puts \\\"hello\\\""; exit\n'),
-                 'r-base:3.6.1': RunArgs(stdin='sprintf("hello")\nq()\n', stdin_sh='R --no-save'),
+                 'jruby:9.3.4': RunArgs(stdin='jruby -e "puts \\\"hello\\\""; exit\n'),
+                 'r-base:4.1.3': RunArgs(stdin='sprintf("hello")\nq()\n', stdin_sh='R --no-save'),
     }
 
-    CMD_ARG = {'perl:5.30': RunArgs(arg='perl -e \'print("hello\\n")\''),
-               'python:3.9': RunArgs(arg='python -c \'print("hello")\''),
-               'pypy:3.5': RunArgs(arg='pypy3 -c \'print("hello")\''),
-               'node:13.13.0': RunArgs(arg='node -e \'console.log("hello")\''),
+    CMD_ARG = {'perl:5.34.1': RunArgs(arg='perl -e \'print("hello\\n")\''),
+               'python:3.10': RunArgs(arg='python -c \'print("hello")\''),
+               'pypy:3.9': RunArgs(arg='pypy3 -c \'print("hello")\''),
+               'node:17.8.0': RunArgs(arg='node -e \'console.log("hello")\''),
     }
 
     # complete listing
     ALL = dict([(b.name, b) for b in
-                [Bench('alpine:3.10.2', 'distro'),
-                 Bench('fedora:30', 'distro'),
-                 Bench('rethinkdb:2.3.6', 'database'),
-                 Bench('postgres:13.1', 'database'),
-                 Bench('redis:5.0.5', 'database'),
-                 Bench('mariadb:10.5', 'database'),
-                 Bench('python:3.9', 'language'),
-                 Bench('golang:1.12.9', 'language'),
-                 Bench('gcc:10.2.0', 'language'),
-                 Bench('jruby:9.2.8.0', 'language'),
-                 Bench('perl:5.30', 'language'),
-                 Bench('php:7.3.8', 'language'),
-                 Bench('pypy:3.5', 'language'),
-                 Bench('r-base:3.6.1', 'language'),
-                 Bench('drupal:8.7.6'),
+                [Bench('alpine:3.15.3', 'distro'),
+                 Bench('fedora:35', 'distro'),
+                 Bench('rethinkdb:2.4.1', 'database'),
+                 Bench('postgres:14.2', 'database'),
+                 Bench('redis:6.2.6', 'database'),
+                 Bench('mariadb:10.7.3', 'database'),
+                 Bench('python:3.10', 'language'),
+                 Bench('golang:1.18', 'language'),
+                 Bench('gcc:11.2.0', 'language'),
+                 Bench('jruby:9.3.4', 'language'),
+                 Bench('perl:5.34.1', 'language'),
+                 Bench('php:8.1.4', 'language'),
+                 Bench('pypy:3.9', 'language'),
+                 Bench('r-base:4.1.3', 'language'),
+                 Bench('drupal:9.3.9'),
                  Bench('jenkins:2.60.3'),
-                 Bench('node:13.13.0'),
-                 Bench('tomcat:10.0.0-jdk15-openjdk-buster', 'web-server'),
-                 Bench('wordpress:5.7', 'web-server'),
+                 Bench('node:17.8.0'),
+                 Bench('tomcat:10.1.0-jdk17-openjdk-bullseye', 'web-server'),
+                 Bench('wordpress:5.9.2', 'web-server'),
+                 Bench('php:8.1.4-apache-bullseye', 'web-server'),
+                 Bench('rabbitmq:3.9.14'),
+                 Bench('elasticsearch:8.1.1'),
+
+                 # needs "--srcrepository=docker.io"
+                 Bench('nixos/nix:2.3.12'),
              ]])
 
-    def __init__(self, repository='docker.io/library', srcrepository='docker.io/library', mode=LEGACY_MODE, optimizer=DEFAULT_OPTIMIZER, puller=DEFAULT_PULLER, pusher=DEFAULT_PUSHER, runtime="containerd"):
+    def __init__(self, repository='docker.io/library', srcrepository='docker.io/library', mode=LEGACY_MODE, optimizer=DEFAULT_OPTIMIZER, puller=DEFAULT_PULLER, pusher=DEFAULT_PUSHER, runtime="containerd", profile=0):
         if runtime == "containerd":
             self.controller = ContainerdController(mode == ESTARGZ_NOOPT_MODE or mode == ESTARGZ_MODE or mode == ZSTDCHUNKED_MODE)
         elif runtime == "podman":
@@ -169,6 +183,10 @@ class BenchRunner:
         self.puller = puller
         self.pusher = pusher
 
+        profile = int(profile)
+        if profile > 0:
+            self.controller.start_profile(profile)
+
     def cleanup(self, cid, bench):
         self.controller.cleanup(cid, self.fully_qualify(bench.name))
 
@@ -297,10 +315,15 @@ class BenchRunner:
         assert(rc == 0)
 
     def convert_cmd_arg_wait(self, src, dest, runargs, option):
+        mounts = ''
+        for a,b in runargs.mount:
+            a = os.path.join(os.path.dirname(os.path.abspath(__file__)), a)
+            a = tmp_copy(a)
+            mounts += '--mount type=bind,src=%s,dst=%s,options=rbind ' % (a,b)
         period = 90
         env = ' '.join(['-env %s=%s' % (k,v) for k,v in runargs.env.iteritems()])
-        cmd = ('%s %s -cni -period %s %s %s %s %s' %
-               (self.optimizer, option, period, env, genargs_for_optimization(runargs.arg), src, dest))
+        cmd = ('%s %s -cni -period %s %s %s %s %s %s' %
+               (self.optimizer, option, period, mounts, env, genargs_for_optimization(runargs.arg), src, dest))
         print cmd
         rc = os.system(cmd)
         assert(rc == 0)
@@ -368,8 +391,8 @@ class BenchRunner:
         src = '%s/%s' % (self.srcrepository, name)
         self.copy_img(src=src, dest=format_repo(LEGACY_MODE, self.repository, name))
         self.convert_and_push_img(src=src, dest=format_repo(ESTARGZ_NOOPT_MODE, self.repository, name))
-        self.optimize_img(name=name, src=src, dest=format_repo(ESTARGZ_MODE, self.repository, name, ""))
-        self.optimize_img(name=name, src=src, dest=format_repo(ZSTDCHUNKED_MODE, self.repository, name, "--zstdchunked"))
+        self.optimize_img(name=name, src=src, dest=format_repo(ESTARGZ_MODE, self.repository, name), option="")
+        self.optimize_img(name=name, src=src, dest=format_repo(ZSTDCHUNKED_MODE, self.repository, name), option="--zstdchunked")
 
     def operation(self, op, bench, cid):
         if op == 'run':
@@ -416,8 +439,12 @@ class ContainerdController:
         if self.is_lazypull:
             snapshotter_opt = "--snapshotter=stargz"
         env = ' '.join(['--env %s=%s' % (k,v) for k,v in runargs.env.iteritems()])
-        cmd = ('%s c create --net-host %s %s -- %s %s %s' %
-               (CTR, snapshotter_opt, env, image, cid, runargs.arg))
+        cmd = '%s c create --net-host %s %s '% (CTR, snapshotter_opt, env)
+        for a,b in runargs.mount:
+            a = os.path.join(os.path.dirname(os.path.abspath(__file__)), a)
+            a = tmp_copy(a)
+            cmd += '--mount type=bind,src=%s,dst=%s,options=rbind ' % (a,b)
+        cmd += '-- %s %s %s' % (image, cid, runargs.arg)
         print cmd
         return cmd
 
@@ -460,10 +487,13 @@ class ContainerdController:
         print cmd
         rc = os.system(cmd)
         assert(rc == 0)
-        cmd = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../reboot_containerd.sh')
-        print cmd
-        rc = os.system(cmd)
-        assert(rc == 0)
+
+    def start_profile(self, seconds):
+        out = open('/tmp/hello-bench-output/profile-%d.out' % (random.randint(1,1000000)), 'w')
+        p = subprocess.Popen(
+            [CTR, 'pprof', '-d', '/run/containerd-stargz-grpc/debug.sock', 'profile', '-s', '%ds' % (seconds)],
+            stdout=out
+        )
 
 class PodmanController:
     def pull_cmd(self, image):
@@ -484,8 +514,12 @@ class PodmanController:
 
     def create_cmd_arg_wait_cmd(self, image, cid, runargs):
         env = ' '.join(['--env %s=%s' % (k,v) for k,v in runargs.env.iteritems()])
-        cmd = ('%s create %s --name %s docker://%s %s ' %
-               (PODMAN, env, cid, image, runargs.arg))
+        cmd = '%s create %s --name %s '% (PODMAN, env, cid)
+        for a,b in runargs.mount:
+            a = os.path.join(os.path.dirname(os.path.abspath(__file__)), a)
+            a = tmp_copy(a)
+            cmd += '--mount type=bind,src=%s,dst=%s ' % (a,b)
+        cmd += ' docker://%s %s ' % (image, runargs.arg)
         print cmd
         return cmd
 
@@ -540,6 +574,7 @@ def main():
         print '--list'
         print '--list-json'
         print '--experiments'
+        print '--profile=<seconds>'
         print '--runtime'
         print '--op=(prepare|run)'
         print '--mode=(%s|%s|%s)' % (LEGACY_MODE, ESTARGZ_NOOPT_MODE, ESTARGZ_MODE, ZSTDCHUNKED_MODE)
@@ -569,6 +604,11 @@ def main():
     trytimes = int(kvargs.pop('experiments', '1'))
     if not op == "run":
         trytimes = 1
+    experiments = range(trytimes)
+
+    profile = int(kvargs.get('profile', 0))
+    if profile > 0:
+        experiments = itertools.count(start=0)
 
     # run benchmarks
     runner = BenchRunner(**kvargs)
@@ -580,7 +620,8 @@ def main():
         create_times = []
         run_times = []
 
-        for i in range(trytimes):
+        for_start = time.time()
+        for i in experiments:
             start = time.time()
             pulltime, createtime, runtime = runner.operation(op, bench, cid)
             elapsed = time.time() - start
@@ -596,6 +637,9 @@ def main():
             print 'create  %s' % createtime
             print 'run     %s' % runtime
 
+            if profile > 0 and time.time() - for_start > profile:
+                break
+
         row = {'mode':'%s' % runner.mode, 'repo':bench.name, 'bench':bench.name, 'elapsed':sum(elapsed_times) / len(elapsed_times), 'elapsed_pull':sum(pull_times) / len(pull_times), 'elapsed_create':sum(create_times) / len(create_times), 'elapsed_run':sum(run_times) / len(run_times)}
         js = json.dumps(row)
         print '%s%s,' % (BENCHMARKOUT_MARK, js)
